/*
 * lxcall-a.g.c
 * Copyright (C) 2011-2014, 2016, 2017, 2019, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * it may be that _post_call assumes things about the transferring function
 * (function_data argument).  originally called by variadics such as `call',
 * `trap', `xcall', it may still guess the said function has the properties of
 * those.
 */

/*
 * the function must copy with synthetic _walk_link outputs set by clients like
 * call_n to intercept the function call.
 */

/*
 * some (most) callers expect a -dispatch- record after a sequenced call.  if
 * the record is to be optimized away (and the function call to be done
 * immediately), callers expectations are to be addressed.
 */

#include <stddef.h>

#include <e4.h>
#include <it.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <lxcall-types.h>
#include <lxcast.h>
#include <track.h>
#include <trans.h>

/*
 * -bits- is a constant.  acknowledging it would cut some code.
 */

#define l_USER(e, output) \
    {									      \
	X1f4_E4_C_USER *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

extern const void *const x1f4_a1_walk_link;

static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(void *, void **,
		     const struct x1f4_function_type *, unsigned,
		     struct x1f4_frame_type *, struct center_type *,
		     struct covert_type *, struct x1f4_linetext_type *);
static int slip_call(void *, void **, void **,
		     const struct x1f4_function_type *, unsigned,
		     struct x1f4_frame_type *, struct center_type *,
		     struct covert_type *, struct x1f4_linetext_type *);

static int
miss_call(void *text, int success, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
pick_call(void *output, void **track,
	  const struct x1f4_function_type *function_data,
	  unsigned bits, struct x1f4_frame_type *frame_data,
	  struct center_type *center_data,
	  struct covert_type *covert_data,
	  struct x1f4_linetext_type *linetext_data)
{
    int status;

    do {
	int type;

	if (linetext_data->function.flags
	    & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    status = _libx1f4i0_lxcall_pick_call
		(track, function_data, 0, bits, frame_data,
		 center_data->track_data, &linetext_data);
	    if (status) {
		break;
	    }
	} else {
	    void **used;

	    used = track[-1];
	    used[X1f4_E4_RENDITION] = &linetext_data->function;
	}

	type = linetext_data->function.type;

	if (type == X1f4_E4_CASE) {
	} else {
	    void *miss;

	    status = frame_data->link
		(frame_data->data, &miss, sizeof(struct x1f4_dxcast_type));
	    if (status) {
		status = _libx1f4i0_lxcall_stat_link(center_data->track_data);
		break;
	    } else {
		struct x1f4_dxcast_type *dxcast_data;

		dxcast_data = miss;

		l_USER(miss, output);

		dxcast_data->call = 0;
		dxcast_data->type = type;

		output = &dxcast_data->data;
	    }
	}

	if (covert_data->scope_data->deck) {
	    status = _libx1f4i0_lxcall_mint_call
		(covert_data, center_data->track_data);
	    if (status) {
		break;
	    }
	}

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);
    } while (0);

    return status;
}


static int
slip_call(void *output, void **input, void **track,
	  const struct x1f4_function_type *function_data,
	  unsigned bits, struct x1f4_frame_type *frame_data,
	  struct center_type *center_data,
	  struct covert_type *covert_data,
	  struct x1f4_linetext_type *linetext_data)
{
    int status;

    do {
	int type;
	unsigned cast, mind, plot, turn;
	void *miss;

	type = linetext_data->function.type;

	if (type == X1f4_E4_CASE) {
	    cast = 0;
	} else {
	    cast = sizeof(struct x1f4_dxcast_type);
	}

	turn = linetext_data->function.flags;

	if (turn & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    if (bits & POST_LINE) {
		mind = 0;
	    } else {
		mind = sizeof(struct x1f4_linetext_type);
	    }
	    if (bits & POST_ARGS) {
	    } else {
		if (turn & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		    mind += (function_data->count - 0) * sizeof(int) << 1;
		} else {
		    if (linetext_data->function.count) {
			mind += sizeof(int) * (function_data->count - 0);
		    }
		}
	    }
	} else {
	    mind = 0;
	}

	if (bits & POST_PLOT) {
	    plot = 0;
	} else {
	    plot = sizeof(struct x1f4_dispatch_type);
	}

	if (cast | mind | plot) {
	    status = frame_data->link
		(frame_data->data, &miss, plot + cast + mind);
	    if (status) {
		status = _libx1f4i0_lxcall_stat_link(center_data->track_data);
		break;
	    }
	}

	{
	    struct x1f4_dispatch_type *dispatch_data;
	    void *subtext;

	    if (bits & POST_PLOT) {
		dispatch_data = input[2];
	    } else {
		dispatch_data = miss;

		miss = dispatch_data + 1;
	    }

	    output = input[0];

	    if (cast) {
		struct x1f4_dxcast_type *dxcast_data;

		dxcast_data = (void *) miss;

		l_USER(dxcast_data, output);

		dxcast_data->call = 0;
		dxcast_data->type = type;

		output = &dxcast_data->data;

		miss = dxcast_data + 1;
	    }

	    subtext = (void *) linetext_data->context;

	    if (turn & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
		status = _libx1f4i0_lxcall_slip_call
		    (track, function_data, 0, bits, &linetext_data, miss,
		     &subtext);
		if (status) {
		    break;
		}
	    } else {
		status = 0;

		((void **) track[-1])[X1f4_E4_RENDITION] =
		    &linetext_data->function;
	    }

	    dispatch_data->back = NULL;
	    dispatch_data->call = miss_call;

	    dispatch_data->transfer.context = subtext;
	    dispatch_data->transfer.function_data = &linetext_data->function;
	    dispatch_data->transfer.input = track;
	    dispatch_data->transfer.output = output;

	    input[2] = dispatch_data;
	    center_data->dispatch_data = dispatch_data;
	}

	if (covert_data->scope_data->deck) {
	    status = _libx1f4i0_lxcall_mint_call
		(covert_data, center_data->track_data);
	}
    } while (0);

    return status;
}


int
_libx1f4i0_lxcall_long_call(struct x1f4_linetext_type *linetext_data,
			    const struct x1f4_function_type *function_data,
			    void *output, void **input,
			    unsigned bits, struct center_type *center_data,
			    struct covert_type *covert_data)
{
    int linked, status;

    do {
	void **track;

	linked = output == x1f4_a1_walk_link;

	if (linked) {
	    track = input[1];
	    center_data->output = input[0];
	    center_data->input = (void *) (input + 1);
	} else {
	    track = input;
	}

	status = _libx1f4i0_lxcall_ever_list
	    (&linetext_data->function, function_data, 0, &track, bits,
	     covert_data->frame_data, center_data);
	if (status) {
	    break;
	}

	if (function_data->flags & X1f4_E4_POST_TYPE) {
	    status = _libx1f4i0_lxcall_suit_line
		(center_data, covert_data, function_data->args,
		 function_data->count, track);
	    if (status) {
		break;
	    }
	}

	if (linked) {
	    status = slip_call
		(output, input, track, function_data, bits,
		 covert_data->frame_data, center_data, covert_data,
		 linetext_data);
	} else {
	    status = pick_call
		(output, track, function_data, bits,
		 covert_data->frame_data, center_data, covert_data,
		 linetext_data);
	}
    } while (0);

    return status;
}
