/*
 * lxtrap-d.c.c
 * Copyright (C) 2019, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <af.h>
#include <e4.h>
#include <e4fine.h>
#include <fd.h>
#include <ln.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxlead.h>
#include <lxtrap-close.h>
#include <lxtrap-drive.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>
#include <tcvset.h>

#define SCALAR_BITS \
    ((1 << X1f4_E4_BILL) | (1 << X1f4_E4_MODE) | (1 << X1f4_E4_REAL))

typedef struct dxlift_type {
    int grip;
    union {
	const struct x1f4_linetext_type *linetext_data;
	const struct x1f4_operator_type *operator_data;
    } lift;
} dxlift_type;

extern const void *const x1f4_a1_walk_link;

static int fare_data(struct fxdata_type *, struct dxlift_type *);
static int line_call(void *, void *, void **);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int near_call(void *, void *, void **);
static int pick_call(void *, void *, void **);
static int slip_call(void *, void *, void **);
static int slip_fall(void *, int, void *, struct x1f4_dispatch_type **);

static int
fare_data(struct fxdata_type *fxdata_data, struct dxlift_type *dxlift_data)
{
    int status;
    struct fxdata_type fxgrip;
    void *grip;

    grip = &fxgrip;
    fxgrip = *fxdata_data;
    if (dxlift_data->grip == 1) {
	const struct x1f4_operator_type *operator_lift;

	operator_lift = dxlift_data->lift.operator_data;
	status = operator_lift->operator(fxdata_data, &grip);
    } else {
	const struct x1f4_linetext_type *linetext_lift;

	linetext_lift = dxlift_data->lift.linetext_data;
	status = linetext_lift->function.function
	    ((void *) linetext_lift->context, fxdata_data, &grip);
    }

    return status;
}


static int
line_call(void *context, void *output, void **input)
{
    int status;

    if (output == x1f4_a1_walk_link) {
	status = slip_call(context, output, input);
    } else {
	status = pick_call(context, output, input);
    }

    return status;
}


static int
miss_call(void *text, int success, void *output,
          struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


/*
 * no improvement in speed from removing the much flags testing.
 * serious sped up from skirting the array (m)allocation.
 * still dismally slow.
 */
static int
near_call(void *context, void *output, void **input)
{
    int status;

    do {
	int mode;
	struct dxlift_type *dxlift_data;
	struct dxnews_type *line_data;
	struct fxdata_type fxdata;
	struct x1f4_linetext_type *linetext_data, *linetext_text;
	unsigned fold, i, mill;
	void *post[16], **track, **used;

	used = input[-1];

	line_data = context;

	dxlift_data = (void *) (line_data + 1);

	mode = *(int *) (dxlift_data + 1);

	linetext_text = line_data->linetext_text;

	fold = linetext_text->function.count;

	track = post;

	*track++ = used;

	input += mode;
	i = fold;
	for (; i; i--) {
	    *track++ = *input++;
	}

	input -= mode + fold;
	track -= fold;

	used[X1f4_E4_RENDITION] = &linetext_text->function;

	status = linetext_text->function.function
	    ((void *) linetext_text->context, &fxdata, track);
	if (status) {
	    break;
	}

	if (dxlift_data->grip) {
	    status = fare_data(&fxdata, dxlift_data);
	    if (status) {
		break;
	    }
	}

	linetext_data = line_data->linetext_data;

	mill = linetext_data->function.count;

	i = mode;
	for (; i; i--) {
	    *track++ = *input++;
	}
	*track++ = &fxdata;
	input += fold;
	i = mill - mode - 1;
	for (; i; i--) {
	    *track++ = *input++;
	}

	track -= mill;

	used[X1f4_E4_RENDITION] = &linetext_data->function;

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);
    } while (0);

    return status;
}


static int
pick_call(void *context, void *output, void **input)
{
    int status;
    void **track = NULL;

    do {
	int mode;
	struct dxlift_type *dxlift_data;
	struct dxnews_type *line_data;
	struct fxdata_type fxdata;
	struct x1f4_function_type function, *function_data;
	struct x1f4_linetext_type *linetext_data, *linetext_text;
	unsigned bits, clip, fold, mill, turn;
	void **used;

	used = input[-1];

	function_data = used[X1f4_E4_RENDITION];

	line_data = context;

	dxlift_data = (void *) (line_data + 1);

	mode = *(int *) (dxlift_data + 1);

	linetext_data = line_data->linetext_data;

	linetext_text = line_data->linetext_text;

	mill = linetext_data->function.count;

	bits = linetext_text->function.flags;

	clip = linetext_data->function.flags;

	fold = linetext_text->function.count;

	if (bits & X1f4_E4_SIDE_LIST) {
	    if (mode + 1 >= mill) {
		fold = function_data->count - mode;
		mill = mode + 1;
	    } else {
		if (clip & X1f4_E4_SIDE_LIST) {
		    mill = function_data->count + 1 - fold;
		}
	    }
	} else {
	    if (clip & X1f4_E4_SIDE_LIST) {
		mill = function_data->count + 1 - fold;
	    }
	}

	if ((bits | clip) & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    turn = sizeof(int);
	    if ((bits | clip) & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		turn <<= 1;
	    }
	} else {
	    turn = 0;
	}

	track = malloc
	    (sizeof(void *)
	     + (fold < mill ? mill : fold) * (sizeof(void *) + turn));
	if (track) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
	    break;
	}

	*track++ = used;

	memcpy(track, input + mode, fold * sizeof(void *));

	if (bits & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    int *args, *very;

	    args = (void *) (track + fold);

	    function = linetext_text->function;

	    function.args = args;

	    very = (int *) function_data->args;

	    memcpy(args, very + mode, fold * sizeof(int));
	    if (bits & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		memcpy(args + fold, very + function_data->count + mode,
		       fold * sizeof(int));
		function.flags |= X1f4_E4_POST_TYPE;
	    }

	    function.count = fold;
	    function.flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	    used[X1f4_E4_RENDITION] = &function;
	} else {
	    used[X1f4_E4_RENDITION] = &linetext_text->function;
	}

	status = linetext_text->function.function
	    ((void *) linetext_text->context, &fxdata, track);
	if (status) {
	    break;
	}

	if (dxlift_data->grip) {
	    status = fare_data(&fxdata, dxlift_data);
	    if (status) {
		break;
	    }
	}

	memcpy(track, input, mode * sizeof(void *));
	track += mode;
	*track = &fxdata;
	memcpy(track + 1, input + mode + fold,
	       (mill - mode - 1) * sizeof(void *));

	track -= mode;

	if (clip & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    int *args, *very;

	    args = (void *) (track + mill);

	    function = linetext_data->function;

	    function.args = args;

	    very = (int *) function_data->args;

	    memcpy(args, very, mode * sizeof(int));
	    args += mode;
	    *args = linetext_text->function.type;
	    memcpy(args + 1, very + mode + fold,
		   (mill - mode - 1) * sizeof(int));
	    if (clip & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		args += mill - mode;
		very += function_data->count;
		memcpy(args, very, mode * sizeof(int));
		args += mode;
		*args++ = 0;
		memcpy(args, very + mode + fold,
		       (mill - mode - 1) * sizeof(int));
		function.flags |= X1f4_E4_POST_TYPE;
	    }

	    function.count = mill;
	    function.flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	    used[X1f4_E4_RENDITION] = &function;
	} else {
	    used[X1f4_E4_RENDITION] = &linetext_data->function;
	}

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);
    } while (0);

    if (track) {
	free(track - 1);
    }

    return status;
}


static int
slip_call(void *context, void *output, void **input)
{
    int status;

    do {
	int mode;
	struct dxlift_type *dxlift_data;
	struct dxnews_type *line_data;
	struct fxdata_type *fxdata_data;
	struct lxtrap_type *lxtrap_data;
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_data, *function_slip;
	struct x1f4_linetext_type *linetext_data, *linetext_text;
	unsigned bits, clip, fold, mill, slip, turn;
	void *patch, **phase, **used, **track;

	phase = input[1];

	used = phase[-1];

	function_data = used[X1f4_E4_RENDITION];

	line_data = context;

	dxlift_data = (void *) (line_data + 1);

	mode = *(int *) (dxlift_data + 1);

	linetext_data = line_data->linetext_data;

	linetext_text = line_data->linetext_text;

	mill = linetext_data->function.count;

	bits = linetext_text->function.flags;

	clip = linetext_data->function.flags;

	fold = linetext_text->function.count;

	if (bits & X1f4_E4_SIDE_LIST) {
	    if (mode + 1 >= mill) {
		fold = function_data->count - mode;
		mill = mode + 1;
	    } else {
		if (clip & X1f4_E4_SIDE_LIST) {
		    mill = function_data->count + 1 - fold;
		}
	    }
	} else {
	    if (clip & X1f4_E4_SIDE_LIST) {
		mill = function_data->count + 1 - fold;
	    }
	}

	lxtrap_data = line_data->lxtrap_data;

	if ((bits | clip) & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    slip = sizeof(struct x1f4_function_type);
	    turn = sizeof(int);
	    if ((bits | clip) & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		turn <<= 1;
	    }
	} else {
	    slip = turn = 0;
	}

	status = lxtrap_data->link_v.link
	    (lxtrap_data->link_v.data, &patch,
	     sizeof(struct x1f4_dispatch_type) + slip + sizeof(void *) * 5
	     + sizeof(struct fxdata_type) + sizeof(void *)
	     + (fold < mill ? mill : fold) * (sizeof(void *) + turn));
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    break;
	}

	dispatch_data = patch;

	fxdata_data = (void *) (dispatch_data + 1);

	track = (void *) (fxdata_data + 1);

	function_slip = (void *) (track + 5);

	track[0] = line_data;
	track[1] = function_data;
	track[2] = phase;
	track[3] = input[0];
	track[4] = (char *) function_slip + slip;

	track = track[4];

	*track++ = used;

	memcpy(track, phase + mode, fold * sizeof(void *));

	if (bits & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    int *args, *very;

	    args = (void *) (track + fold);

	    *function_slip = linetext_text->function;

	    function_slip->args = args;

	    very = (int *) function_data->args;

	    memcpy(args, very + mode, fold * sizeof(int));
	    if (bits & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		memcpy(args + fold, very + function_data->count + mode,
		       fold * sizeof(int));
		function_slip->flags |= X1f4_E4_POST_TYPE;
	    }

	    function_slip->count = fold;
	    function_slip->flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	    used[X1f4_E4_RENDITION] = function_slip;
	} else {
	    used[X1f4_E4_RENDITION] = &linetext_text->function;
	}

	dispatch_data->back = patch;
	dispatch_data->call = slip_fall;

	dispatch_data->transfer.context = (void *) linetext_text->context;
	dispatch_data->transfer.function_data = &linetext_text->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = fxdata_data;

	input[2] = dispatch_data;
    } while (0);

    return status;
}


static int
slip_fall(void *text, int status, void *output,
          struct x1f4_dispatch_type **dispatch)
{
    do {
	int mode;
	struct dxlift_type *dxlift_data;
	struct dxnews_type *line_data;
	struct fxdata_type *fxdata_data;
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_data;
	struct x1f4_linetext_type *linetext_data, *linetext_text;
	unsigned bits, clip, fold, mill;
	void *output, **phase, **used, **track;

	if (status) {
	    break;
	}

	dispatch_data = text;

	fxdata_data = (void *) (dispatch_data + 1);

	track = (void *) (fxdata_data + 1);

	line_data = *track;
	function_data = track[1];
	phase = track[2];
	output = track[3];

	track = track[4];

	used = phase[-1];

	dxlift_data = (void *) (line_data + 1);

	mode = *(int *) (dxlift_data + 1);

	if (dxlift_data->grip) {
	    status = fare_data(fxdata_data, dxlift_data);
	    if (status) {
		break;
	    }
	}

	linetext_data = line_data->linetext_data;

	linetext_text = line_data->linetext_text;

	mill = linetext_data->function.count;

	bits = linetext_text->function.flags;

	clip = linetext_data->function.flags;

	fold = linetext_text->function.count;

	if (bits & X1f4_E4_SIDE_LIST) {
	    if (mode + 1 >= mill) {
		fold = function_data->count - mode;
		mill = mode + 1;
	    } else {
		if (clip & X1f4_E4_SIDE_LIST) {
		    mill = function_data->count + 1 - fold;
		}
	    }
	} else {
	    if (clip & X1f4_E4_SIDE_LIST) {
		mill = function_data->count + 1 - fold;
	    }
	}

	*track++ = used;

	memcpy(track, phase, mode * sizeof(void *));
	track += mode;
	*track = fxdata_data;
	memcpy(track + 1, phase + mode + fold,
	       (mill - mode - 1) * sizeof(void *));

	track -= mode;

	if (clip & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    int *args, *very;
	    struct x1f4_function_type *function_slip;

	    args = (void *) (track + mill);

	    function_slip = (void *) (track - 1);
	    function_slip--;

	    *function_slip = linetext_data->function;

	    function_slip->args = args;

	    very = (int *) function_data->args;

	    memcpy(args, very, mode * sizeof(int));
	    args += mode;
	    *args = linetext_text->function.type;
	    memcpy(args + 1, very + mode + fold,
		   (mill - mode - 1) * sizeof(int));
	    if (clip & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
		args += mill - mode;
		very += function_data->count;
		memcpy(args, very, mode * sizeof(int));
		args += mode;
		*args++ = 0;
		memcpy(args, very + mode + fold,
		       (mill - mode - 1) * sizeof(int));
		function_slip->flags |= X1f4_E4_POST_TYPE;
	    }

	    function_slip->count = mill;
	    function_slip->flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	    used[X1f4_E4_RENDITION] = function_slip;
	} else {
	    used[X1f4_E4_RENDITION] = &linetext_data->function;
	}

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = output;

	*dispatch = dispatch_data;
    } while (0);

    return status;
}


int
_libx1f4i0_lxtrap_c_derived(void *context, void *output, void **input)
{
    int linked = 0, status;
    struct dxnews_type *line_data = NULL;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_leadnode_type *leadnode_data = NULL;
    struct x1f4_linetext_type *linetext_data, *linetext_text;

    lxtrap_data = context;

    do {
	X1f4_E4_C_MODE mode;
	int *args, (*fare)(void *, void *, void **), type;
	const int *call, *such;
	struct dxlift_type dxlift, *dxlift_data;
	struct x1f4_linetext_type *linetext_over;
	struct x1f4_nodetype_type *nodetype_data;
	unsigned bits, fold, late, norm, mill, over, post;
	void *line;

	linetext_data = I_USER(&I_CASE(input[0])->data);

	mill = linetext_data->function.count;

	over = linetext_data->function.flags;

	mode = I_MODE(input[1]);

	if (over & X1f4_E4_SIDE_LIST) {
	    if (mode < 0) {
		status = _libx1f4i0_lxtrap_stat_slip(context, mill, mode);
		if (1) {
		    break;
		}
	    }
	} else {
	    status = _libx1f4i0_lxtrap_mill_lock(context, &mode, mill);
	    if (status) {
		break;
	    }
	}

	linetext_text = I_USER(&I_CASE(input[2])->data);

	such = linetext_data->function.args;

	norm = linetext_text->function.flags;

	bits = norm | over;

	dxlift.grip = 0;

	status = 0;

	if (mode < mill) {
	    int half;

	    half = such[mode];
	    if (half != X1f4_E4_SLIP) {
		int name;

		name = linetext_text->function.type;
		if (half != name) {
		    do {
			if (lxtrap_data->link_p.link) {
			    if (lxtrap_data->link_p.link
				(lxtrap_data->link_p.text, half, name,
				 &dxlift.lift.operator_data)) {
			    } else {
				dxlift.grip = 1;
				break;
			    }
			}

			if (lxtrap_data->link_q.link) {
			    if (lxtrap_data->link_q.link
				(lxtrap_data->link_q.text, half, name,
				 &dxlift.lift.linetext_data)) {
			    } else {
				dxlift.grip = 2;
				break;
			    }
			}

			status = _libx1f4i0_lxcall_stat_lock
			    (&lxtrap_data->link_e,
			     lxtrap_data->trap_e.eelookup_data,
			     linetext_text->function.type, half, mode,
			     &linetext_data->function);
		    } while (0);
		    if (status) {
			break;
		    }
		}
	    }

	    if (over & X1f4_E4_POST_TYPE) {
		const int *pass;
		unsigned j = 0;

		pass = such + mill;

		if (pass[mode] & X1f4_E4_POST_XSET) {
#if __LEFT_IMMEDIATE_REFERENCES__
		    if (half < 16 && (1 << half) & SCALAR_BITS
			&& !dxlift.grip) {
		    } else {
#endif				/* __LEFT_IMMEDIATE_REFERENCES__ */
			status = _libx1f4i0_lxcall_stat_fink
			    (&lxtrap_data->link_e, mode,
			     &linetext_data->function);
			if (1) {
			    break;
			}
#if __LEFT_IMMEDIATE_REFERENCES__
		    }
#endif				/* __LEFT_IMMEDIATE_REFERENCES__ */
		}

		for (; j < mill; j++, pass++) {
		    if (j != mode && *pass) {
			break;
		    }
		}
		if (j < mill) {
		} else {
		    over &= ~X1f4_E4_POST_TYPE;
		}
	    }

	    if (norm & X1f4_E4_SIDE_LIST) {
		if (mode < mill - 1) {
		    norm &= ~X1f4_E4_SIDE_LIST;
		}
	    }
	}

	fold = linetext_text->function.count;

	if (mode < mill) {
	    late = mill + fold - 1;
	} else {
	    late = mode + fold;
	}

	post = (norm | over) & X1f4_E4_POST_TYPE ? 1 : 0;

	status = lxtrap_data->link_m.link
	    (lxtrap_data->link_m.data, &line,
	     sizeof(struct x1f4_leadnode_type)
	     + sizeof(struct x1f4_linetext_type)
	     + sizeof(struct dxlift_type) + sizeof(struct dxnews_type)
	     + sizeof(int) * (1 + (late << post)));
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	leadnode_data = line;

	linetext_over = (void *) (leadnode_data + 1);

	line_data = (void *) (linetext_over + 1);

	leadnode_data->line.call = 0;

	_libx1f4i0_init_tcvset
	    (&leadnode_data->line.pset, &leadnode_data->line.text);

	line_data->lxtrap_data = lxtrap_data;

	dxlift_data = (void *) (line_data + 1);

	args = (int *) (dxlift_data + 1);

	*dxlift_data = dxlift;

	*args++ = mode;

	call = linetext_text->function.args;

	linetext_over->function.args = args;

	if (mode < mill) {
	    unsigned past;

	    past = mill - 1 - mode;

	    memcpy(args, such, mode * sizeof(int));
	    args += mode;
	    memcpy(args, call, fold * sizeof(int));
	    args += fold;
	    such += mode + 1;
	    memcpy(args, such, past * sizeof(int));
	    if (post) {
		unsigned i;

		args += past;
		if (over & X1f4_E4_POST_TYPE) {
		    such += past;
		    for (i = mode; i; i--) {
			*args++ = *such++ & X1f4_E4_POST_XSET;
		    }
		} else {
		    for (i = mode; i; i--) {
			*args++ = 0;
		    }
		}
		if (norm & X1f4_E4_POST_TYPE) {
		    call += fold;
		    for (i = fold; i; i--) {
			*args++ = *call++ & X1f4_E4_POST_XSET;
		    }
		} else {
		    for (i = fold; i; i--) {
			*args++ = 0;
		    }
		}
		if (over & X1f4_E4_POST_TYPE) {
		    for (i = past; i; i--) {
			*args++ = *++such & X1f4_E4_POST_XSET;
		    }
		} else {
		    for (i = past; i; i--) {
			*args++ = 0;
		    }
		}
	    }
	} else {
	    unsigned i;

	    memcpy(args, such, mill * sizeof(int));
	    args += mill;
	    for (i = mode - mill; i; i--) {
		*args++ = X1f4_E4_SLIP;
	    }
	    memcpy(args, call, fold * sizeof(int));
	    if (post) {
		args += fold;
		if (over & X1f4_E4_POST_TYPE) {
		    such += mill;
		    for (i = mill; i; i--) {
			*args++ = *such++ & X1f4_E4_POST_XSET;
		    }
		} else {
		    for (i = mill; i; i--) {
			*args++ = 0;
		    }
		}
		if (1) {
		    for (i = mode - mill; i; i--) {
			*args++ = 0;
		    }
		}
		if (norm & X1f4_E4_POST_TYPE) {
		    call += fold;
		    for (i = fold; i; i--) {
			*args++ = *call++ & X1f4_E4_POST_XSET;
		    }
		} else {
		    for (i = fold; i; i--) {
			*args++ = 0;
		    }
		}
	    }
	}

	fare = line_call;

	linetext_over->context = line_data;

	linetext_over->function.count = late;
	linetext_over->function.length = 6;
	linetext_over->function.name = "consed";
	linetext_over->function.type = linetext_data->function.type;

	linetext_over->function.flags = X1f4_E4_OVER_LINK | X1f4_E4_TEXT_LINK
	    | ((norm | over)
	       & ~(X1f4_E4_LEAD_CALL | X1f4_E4_NEWS_LINK | X1f4_E4_NODE_LINK));
	if ((norm | over) & (X1f4_E4_LEAD_CALL | X1f4_E4_NODE_LINK)) {
	    linetext_over->function.flags |= X1f4_E4_WALK_LINK;
	}

	if (linetext_over->function.flags & X1f4_E4_WALK_LINK) {
	} else {
	    if (bits & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)
		| (mill | fold) & ~15) {
		fare = pick_call;
	    } else {
		fare = near_call;
	    }
	}

	linetext_over->function.function = fare;

	nodetype_data = lxtrap_data->link_y.fold[0].nodetype;

	leadnode_data->link.nodetype_data = nodetype_data;
	leadnode_data->link.slip = NULL;

	line_data->track = NULL;

	status = _x1f4_node_effect
	    (NULL, (void *) &line_data->linetext_data, linetext_data,
	     &leadnode_data->link);
	if (status) {
	    break;
	}

	linked = 1;

	status = _x1f4_node_effect
	    (NULL, (void *) &line_data->linetext_text, linetext_text,
	     &leadnode_data->link);
	if (status) {
	    break;
	}

	linked = 2;

	line_data->fold = 0;

	status = x1f4_type_e4fine
	    (lxtrap_data->link_o.fine, &type, linetext_over);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	status = x1f4_cast_node
	    (&lxtrap_data->link_v, &lxtrap_data->link_e, nodetype_data, type,
	     &linetext_over, output);
    } while (0);

    if (status) {
	if (leadnode_data) {
	    if (linked) {
		if (1 < linked) {
		    _x1f4_free_effect
			(NULL, linetext_text, NULL, &leadnode_data->link);
		}

		_x1f4_free_effect
		    (NULL, linetext_data, NULL, &leadnode_data->link);
	    }

	    lxtrap_data->link_m.free
		(lxtrap_data->link_m.data, leadnode_data);
	}
    }

    return status;
}
