/*
 * lxtrap-i.i.c
 * Copyright (C) 2019-2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stddef.h>
#include <stdlib.h>

#include <af.h>
#include <e4.h>
#include <e4fine.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxtrap-drive.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>
#include <tcvset.h>

#define SCALAR_BITS \
    ((1 << X1f4_E4_BILL) | (1 << X1f4_E4_MODE) | (1 << X1f4_E4_REAL))

extern const void *const x1f4_a1_walk_link;

static int join_call(void **, void **, unsigned, unsigned,
		     struct dxnews_type *, struct x1f4_function_type *);
static int join_fall(void **, int *, void **, const int *, unsigned, unsigned,
		     unsigned, struct dxnews_type *);
static int line_call(void *, void *, void **);
static int line_fall(void *, void *, void **);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(void *, void *, void **);
static int pick_fall(void *, void *, void **);
static int slip_call(void *, void *, void **);
static int slip_fall(void *, void *, void **);

static int
join_call(void **track, void **input, unsigned fold, unsigned late,
	  struct dxnews_type *line_data,
	  struct x1f4_function_type *function_data)
{
    unsigned i, j;
    unsigned char *fuse;
    void **pitch, **used;

    *track++ = used = input[-1];

    used[X1f4_E4_RENDITION] = function_data;

    pitch = line_data->track;

    fuse = line_data->fuse;

    for (i = fold; i; i--) {
	j = *fuse++;
	late -= j;
	for (; j; j--) {
	    *track++ = *input++;
	}
	{
	    *track++ = *pitch++;
	}
    }
    {
	for (j = late; j; j--) {
	    *track++ = *input++;
	}
    }

    return 0;
}


static int
join_fall(void **track, int *args, void **input, const int *fill,
	  unsigned fold, unsigned late, unsigned post,
	  struct dxnews_type *line_data)
{
    const int *wind;
    int *jack;
    unsigned i, j;
    unsigned char *fuse;
    void **pitch;

    *track++ = input[-1];

    pitch = line_data->track;

    fuse = line_data->fuse;

    jack = (void *) (line_data + 1);

    wind = fill + late;

    if (post) {
	post = fold + late;
    }

    for (i = fold; i; i--) {
	j = *fuse++;
	late -= j;
	for (; j; j--) {
	    *track++ = *input++;
	    if (post) {
		args[post] = *wind++;
	    }
	    *args++ = *fill++;
	}
	if (post) {
	    args[post] = 0;
	}
	{
	    *track++ = *pitch++;
	    *args++ = *jack++;
	}
    }
    if (post) {
	for (j = late; j; j--) {
	    *track++ = *input++;
	    args[post] = *wind++;
	    *args++ = *fill++;
	}
    } else {
	for (j = late; j; j--) {
	    *track++ = *input++;
	    *args++ = *fill++;
	}
    }

    return 0;
}


static int
line_call(void *context, void *output, void **input)
{
    return output == x1f4_a1_walk_link
	? slip_call(context, output, input)
	: pick_call(context, output, input);
}


static int
line_fall(void *context, void *output, void **input)
{
    return output == x1f4_a1_walk_link
	? slip_fall(context, output, input)
	: pick_fall(context, output, input);
}


static int
miss_call(void *text, int success, void *output,
          struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
pick_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late;
    void **track;

    line_data = context;

    linetext_data = line_data->linetext_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    track = malloc((1 + fold + late) * sizeof(void *));
    if (track) {
	join_call
	    (track, input, fold, late, line_data, &linetext_data->function);

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track + 1);

	free(track);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
pick_fall(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void **used, **track;

    used = input[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    track = malloc((1 + size) * sizeof(void *) + (size * sizeof(int) << post));
    if (track) {
	int *args;
	struct x1f4_function_type function;

	args = (void *) (track + 1 + size);

	join_fall
	    (track, args, input, function_data->args, fold, late, post,
	     line_data);

	function = linetext_data->function;

	function.args = args;
	function.count = size;
	function.flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	if (post) {
	    function.flags |= X1f4_E4_POST_TYPE;
	}

	used[X1f4_E4_RENDITION] = &function;

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track + 1);

	free(track);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
slip_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    unsigned fold, late;
    void *patch;

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type)
	 + (fold + late + 1) * sizeof(void *));
    if (status) {
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_linetext_type *linetext_data;
	void **track;

	linetext_data = line_data->linetext_data;

	dispatch_data = patch;

	track = (void *) (dispatch_data + 1);

	join_call
	    (track, input[1], fold, late, line_data, &linetext_data->function);

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track + 1;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


static int
slip_fall(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void *patch, **phase, **used;

    phase = input[1];

    used = phase[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + sizeof(struct x1f4_function_type)
	 + (size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (status) {
    } else {
	int *args;
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_slip;
	void **track;

	dispatch_data = patch;

	function_slip = (void *) (dispatch_data + 1);

	track = (void *) (function_slip + 1);

	args = (void *) (track + 1 + size);

	join_fall
	    (track, args, phase, function_data->args, fold, late, post,
	     line_data);

	*function_slip = linetext_data->function;

	function_slip->args = args;
	function_slip->count = size;
	function_slip->flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	if (post) {
	    function_slip->flags |= X1f4_E4_POST_TYPE;
	}

	used[X1f4_E4_RENDITION] = function_slip;

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track + 1;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


int
_libx1f4i0_lxtrap_half_into(void *context, void *output, void **input,
			    struct x1f4_linetext_type *linetext_data,
			    struct x1f4_function_type *function_data,
			    unsigned fold, unsigned late, unsigned fast,
			    unsigned char **yank)
{
    int linked = 0, status = 0;
    struct dxnews_type *line_data = NULL;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_leadnode_type *leadnode_data = NULL;

    lxtrap_data = context;

    do {
	int *args, *edge, *suit, type;
	const int *fill, *such;
	struct frame_type frame;
	struct x1f4_function_type function[2];
	struct x1f4_linetext_type *linetext_over;
	struct x1f4_nodetype_type *nodetype_data;
	unsigned bits = 0, i, none = X1f4_E4_SLIP_LIST, over, post, rise, till;
	unsigned char *fuse, *rule;
	void *line, *spin, **track;

	over = linetext_data->function.flags;

	till = linetext_data->function.count;

	such = linetext_data->function.args;

	post = over & X1f4_E4_POST_TYPE ? 1 : 0;

	rule = *yank;

	if (post) {
	    const int *pass;
	    int less, rest = 0;

	    pass = such + till;
	    i = 0;
	    for (; i < till; i++, pass++) {
		less = *pass;
		if (less) {
		    if (rule[i]) {
			if (less & X1f4_E4_POST_XSET) {
#if __LEFT_IMMEDIATE_REFERENCES__
			    int half;

			    half = *(pass - till);
			    if (half < 16 && (1 << half) & SCALAR_BITS) {
			    } else {
#endif				/* __LEFT_IMMEDIATE_REFERENCES__ */
				status = _libx1f4i0_lxcall_stat_fink
				    (&lxtrap_data->link_e, i,
				     &linetext_data->function);
				if (1) {
				    break;
				}
#if __LEFT_IMMEDIATE_REFERENCES__
			    }
#endif				/* __LEFT_IMMEDIATE_REFERENCES__ */
			}
		    } else {
			rest = 1;
		    }
		}
	    }
	    if (status) {
		break;
	    }
	    if (rest) {
	    } else {
		post = 0;
		over &= ~X1f4_E4_POST_TYPE;
	    }
	}

	status = lxtrap_data->link_m.link
	    (lxtrap_data->link_m.data, &line,
	     sizeof(struct x1f4_leadnode_type)
	     + sizeof(struct x1f4_linetext_type)
	     + sizeof(dxnews_type) + (sizeof(int) * (fold + (late << post)))
	     + fold);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	leadnode_data = line;

	linetext_over = (void *) (leadnode_data + 1);

	line = linetext_over + 1;

	line_data = line;

	leadnode_data->line.call = 0;

	_libx1f4i0_init_tcvset
	    (&leadnode_data->line.pset, &leadnode_data->line.text);

	line_data->lxtrap_data = lxtrap_data;

	args = (void *) (line_data + 1);

	suit = args + fold;

	fuse = (void *) (suit + (late << post));

	line_data->fuse = fuse;

	line_data->track = NULL;

	track = spin = malloc((sizeof(void *) + sizeof(int)) * fold);
	if (track) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	fill = function_data->args;

	edge = (void *) (track + fold);

	rise = 0;
	for (i = 0; i < fast; i++) {
	    unsigned j;

	    if (i < till) {
		type = *such;
	    } else {
		type = X1f4_E4_SLIP;
	    }
	    j = *rule++;
	    if (j) {
		if (type == X1f4_E4_SLIP) {
		    type = function_data->args[j];
		}

		*args++ = type;
		*edge++ = fill[j];
		*track++ = input[j];
		*fuse++ = rise;
		rise = 0;
	    } else {
		if (i < till) {
		    *suit = type;
		    if (post) {
			suit[late] = such[till];
		    }
		} else {
		    bits = X1f4_E4_SLIP_LIST;
		    *suit = type;
		    if (post) {
			suit[late] =
			    over & X1f4_E4_LINK_PASS ? X1f4_E4_POST_XSET : 0;
		    }
		}
		suit++;
		rise++;
		if (type == X1f4_E4_SLIP) {
		    none = 0;
		}
	    }

	    such++;
	}

	free(rule - fast);

	*yank = NULL;

	args -= fold;

	track -= fold;

	linetext_over->context = line_data;

	linetext_over->function.args = args + fold;
	linetext_over->function.count = late;
	linetext_over->function.function =
	    over & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)
	    ? line_fall : line_call;
	linetext_over->function.length = 7;
	linetext_over->function.name = "applied";
	linetext_over->function.type = linetext_data->function.type;

	linetext_over->function.flags = bits | X1f4_E4_OVER_LINK
	    | X1f4_E4_TEXT_LINK
	    | (over & ~(X1f4_E4_LEAD_CALL | X1f4_E4_NEWS_LINK
			| X1f4_E4_NODE_LINK));
	linetext_over->function.flags &= ~none;

	/*
	 * setting _WALK_LINK bit for reasons...
	 */
	if (over & (X1f4_E4_LEAD_CALL | X1f4_E4_NODE_LINK) || 1) {
	    linetext_over->function.flags |= X1f4_E4_WALK_LINK;
	}

	nodetype_data = lxtrap_data->link_y.fold[0].nodetype;

	leadnode_data->link.nodetype_data = nodetype_data;
	leadnode_data->link.slip = NULL;

	frame.data = lxtrap_data->link_m.data;
	frame.link = lxtrap_data->link_m.link;

	function[0] = linetext_data->function;
	function[0].args = args;
	function[0].flags &= ~X1f4_E4_POST_TYPE;
	function[0].count = fold;

	function[1] = linetext_data->function;
	function[1].args = edge - fold;
	function[1].flags &= ~X1f4_E4_POST_TYPE;
	function[1].count = fold;

	status = _libx1f4i0_lxcall_even_list
	    (function, function + 1, 0, &track,
	     X1f4_LXCALL_POST_COPY | X1f4_LXCALL_POST_HOLD
	     /* | X1f4_LXCALL_POST_WAYS */,
	     &frame, &lxtrap_data->link_p, &lxtrap_data->link_q,
	     &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data);

	free(spin);

	spin = NULL;

	if (status) {
	    break;
	}

	line_data->track = track;

	status = _x1f4_node_effect
	    (NULL, (void *) &line_data->linetext_data, linetext_data,
	     &leadnode_data->link);
	if (status) {
	    break;
	}

	linked = 1;

	line_data->linetext_text = NULL;

	status = _libx1f4i0_lxtrap_node_args
	    (line_data, args, fold, &leadnode_data->link);
	if (status) {
	    if (1) {
		break;
	    }
	}

	status = x1f4_type_e4fine
	    (lxtrap_data->link_o.fine, &type, linetext_over);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	status = x1f4_cast_node
	    (&lxtrap_data->link_v, &lxtrap_data->link_e, nodetype_data, type,
	     &linetext_over, output);
    } while (0);

    if (status) {
	if (leadnode_data) {
	    if (line_data->track) {
		if (linked) {
		    if (line_data->fold) {
			_libx1f4i0_lxtrap_free_args(line_data, NULL);
		    }

		    _x1f4_free_effect
			(NULL, linetext_data, NULL, &leadnode_data->link);
		}

		lxtrap_data->link_m.free
		    (lxtrap_data->link_m.data, line_data->track - 1);
	    }

	    lxtrap_data->link_m.free
		(lxtrap_data->link_m.data, leadnode_data);
	}
    }

    return status;
}
