/*
 * lxtrap-l.a.c
 * Copyright (C) 2019, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <af.h>
#include <e4.h>
#include <e4fine.h>
#include <ln.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxlead.h>
#include <lxtrap-close.h>
#include <lxtrap-drive.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>
#include <tcvset.h>

#define SCALAR_BITS \
    ((1 << X1f4_E4_BILL) | (1 << X1f4_E4_MODE) | (1 << X1f4_E4_REAL))

extern const void *const x1f4_a1_walk_link;

static int line_call(void *, void *, void **);
static int line_fall(void *, void *, void **);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(void *, void *, void **);
static int pick_fall(void *, void *, void **);
static int slip_call(void *, void *, void **);
static int slip_fall(void *, void *, void **);

static int
line_call(void *context, void *output, void **input)
{
    return output == x1f4_a1_walk_link
	? slip_call(context, output, input)
	: pick_call(context, output, input);
}


static int
line_fall(void *context, void *output, void **input)
{
    return output == x1f4_a1_walk_link
	? slip_fall(context, output, input)
	: pick_fall(context, output, input);
}


static int
miss_call(void *text, int success, void *output,
          struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
pick_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late;
    void **track;

    line_data = context;

    linetext_data = line_data->linetext_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    track = malloc((1 + fold + late) * sizeof(void *));
    if (track) {
	void **used;

	*track++ = used = input[-1];

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, input, late * sizeof(void *));

	used[X1f4_E4_RENDITION] = &linetext_data->function;

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);

	free(track - 1);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
pick_fall(void *context, void *output, void **input)
{
    int *args, status;
    struct dxnews_type *line_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void **used, **track;

    used = input[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    track = malloc
	((size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (track) {
	struct x1f4_function_type function;

	*track++ = input[-1];

	args = (void *) (track + size);

	function = linetext_data->function;

	memcpy(args, line_data + 1, fold * sizeof(int));
	memcpy(args + fold, function_data->args, late * sizeof(int));
	if (post) {
	    int *very;

	    very = args + size;
	    memset(very, 0, fold * sizeof(int));
	    memcpy(very + fold, function_data->args + late,
		   late * sizeof(int));

	    function.flags |= X1f4_E4_POST_TYPE;
	}

	function.args = args;
	function.count = size;
	function.flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	used[X1f4_E4_RENDITION] = &function;

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, input, late * sizeof(void *));

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track);

	free(track - 1);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
slip_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    unsigned fold, late, size;
    void *patch;

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + (size + 1) * sizeof(void *));
    if (status) {
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_linetext_type *linetext_data;
	void **phase, **track, **used;

	linetext_data = line_data->linetext_data;

	dispatch_data = patch;

	track = (void *) (dispatch_data + 1);

	phase = input[1];

	*track++ = used = phase[-1];

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, phase, late * sizeof(void *));

	used[X1f4_E4_RENDITION] = &linetext_data->function;

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


static int
slip_fall(void *context, void *output, void **input)
{
    int *args, status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void *patch, **phase, **used;

    phase = input[1];

    used = phase[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + sizeof(struct x1f4_function_type)
	 + (size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (status) {
	status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_slip;
	void **track;

	dispatch_data = patch;

	function_slip = (void *) (dispatch_data + 1);

	track = (void *) (function_slip + 1);

	*track++ = used;

	args = (void *) (track + size);

	*function_slip = linetext_data->function;

	memcpy(args, line_data + 1, fold * sizeof(int));
	memcpy(args + fold, function_data->args, late * sizeof(int));
	if (post) {
	    int *very;

	    very = args + size;
	    memset(very, 0, fold * sizeof(int));
	    memcpy(very + fold, function_data->args + late,
		   late * sizeof(int));

	    function_slip->flags |= X1f4_E4_POST_TYPE;
	}

	function_slip->args = args;
	function_slip->count = size;
	function_slip->flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	used[X1f4_E4_RENDITION] = function_slip;

	memcpy(track, line_data->track, fold * sizeof(void *));
	memcpy(track + fold, phase, late * sizeof(void *));

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


int
_libx1f4i0_lxtrap_a_logique(void *context, void *output, void **input)
{
    int linked = 0, status;
    struct dxnews_type *line_data = NULL;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_leadnode_type *leadnode_data = NULL;
    struct x1f4_linetext_type *linetext_data;

    lxtrap_data = context;

    do {
	int *args, type;
	const int *such;
	struct frame_type frame;
	struct x1f4_dxcast_type *dxcast_data;
	struct x1f4_function_type function, *function_data, *function_text;
	struct x1f4_linetext_type *linetext_over;
	struct x1f4_nodetype_type *nodetype_data;
	unsigned fold, late, mill, norm, over, post;
	void *line, **track, **used;

	dxcast_data = I_USER(input[0]);

	linetext_data = I_USER(&dxcast_data->data);

	used = input[-1];

	function_data = used[X1f4_E4_RENDITION];

	fold = function_data->count - 1;

	if (fold) {
	} else {
	    l_USER(dxcast_data, output);

	    status = 0;

	    break;
	}

	over = linetext_data->function.flags;

	mill = linetext_data->function.count;

	late = mill - fold;
	if ((int) late < 0) {
	    late = 0;
	    norm = mill;
	} else {
	    norm = fold;
	}

	such = linetext_data->function.args;

	post = over & X1f4_E4_POST_TYPE ? 1 : 0;

	if (post) {
	    const int *pass;
	    unsigned j;

	    pass = such + mill;
	    j = norm;
	    for (; j; j--) {
		if (*pass++ & X1f4_E4_POST_XSET) {
#if __LEFT_IMMEDIATE_REFERENCES__
		    int half;

		    half = *(pass - mill);
		    if (half < 16 && (1 << half) & SCALAR_BITS) {
		    } else {
#endif				/* __LEFT_IMMEDIATE_REFERENCES__ */
			status = _libx1f4i0_lxcall_stat_fink
			    (&lxtrap_data->link_e, norm - j,
			     &linetext_data->function);
			if (1) {
			    break;
			}
#if __LEFT_IMMEDIATE_REFERENCES__
		    }
#endif				/* __LEFT_IMMEDIATE_REFERENCES__ */
		}
	    }
	    if (j) {
		break;
	    }
	    j = mill - norm;
	    for (; j; j--) {
		if (*pass++) {
		    break;
		}
	    }
	    if (j) {
	    } else {
		post = 0;
		over &= ~X1f4_E4_POST_TYPE;
	    }
	}

	status = lxtrap_data->link_m.link
	    (lxtrap_data->link_m.data, &line,
	     sizeof(struct x1f4_leadnode_type)
	     + sizeof(struct x1f4_linetext_type)
	     + sizeof(dxnews_type) + (sizeof(int) * (fold + (late << post))));
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	leadnode_data = line;

	linetext_over = (void *) (leadnode_data + 1);

	line = linetext_over + 1;

	line_data = line;

	leadnode_data->line.call = 0;

	_libx1f4i0_init_tcvset
	    (&leadnode_data->line.pset, &leadnode_data->line.text);

	line_data->lxtrap_data = lxtrap_data;

	args = (int *) (line_data + 1);

	memcpy(args, such, norm * sizeof(int));
	memcpy(args + norm, function_data->args + 1 + norm,
	       (fold - norm) * sizeof(int));

	memcpy(args + fold, such + norm, late * sizeof(int));
	if (post) {
	    memcpy(args + fold + late, such + mill + norm, late * sizeof(int));
	}

	if (over & X1f4_E4_SLIP_LIST) {
	    unsigned j;

	    j = norm;
	    for (; j; j--, args++) {
		if (*args == X1f4_E4_SLIP) {
		    *args = function_data->args[norm - j + 1];
		}
	    }

	    args -= norm;
	}

	linetext_over->context = line_data;

	linetext_over->function.args = args + fold;
	linetext_over->function.count = late;
	linetext_over->function.function =
	    over & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)
	    ? line_fall : line_call;
	linetext_over->function.length = 7;
	linetext_over->function.name = "applied";
	linetext_over->function.type = linetext_data->function.type;

	linetext_over->function.flags = X1f4_E4_OVER_LINK | X1f4_E4_TEXT_LINK
	    | (over & ~(X1f4_E4_LEAD_CALL | X1f4_E4_NEWS_LINK
			| X1f4_E4_NODE_LINK));
	if (over & (X1f4_E4_LEAD_CALL | X1f4_E4_NODE_LINK)) {
	    linetext_over->function.flags |= X1f4_E4_WALK_LINK;
	}

	nodetype_data = lxtrap_data->link_y.fold[0].nodetype;

	leadnode_data->link.nodetype_data = nodetype_data;
	leadnode_data->link.slip = NULL;

	frame.data = lxtrap_data->link_m.data;
	frame.link = lxtrap_data->link_m.link;

	track = input + 1;

	function_text = &linetext_data->function;

	if (late) {
	    function = *function_text;
	    function_text = &function;
	    function.flags &= ~X1f4_E4_POST_TYPE;
	    function.count = fold;
	}

	status = _libx1f4i0_lxcall_even_list
	    (function_text, function_data, 1, &track,
	     X1f4_LXCALL_POST_COPY /* | X1f4_LXCALL_POST_WAYS */,
	     &frame, &lxtrap_data->link_p, &lxtrap_data->link_q,
	     &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data);

	if (status) {
	    line_data->track = NULL;

	    if (1) {
		break;
	    }
	}

	line_data->track = track;

	status = _x1f4_node_effect
	    (NULL, (void *) &line_data->linetext_data, linetext_data,
	     &leadnode_data->link);
	if (status) {
	    break;
	}

	linked = 1;

	line_data->linetext_text = NULL;

	status = _libx1f4i0_lxtrap_node_args
	    (line_data, args, fold, &leadnode_data->link);
	if (status) {
	    if (1) {
		break;
	    }
	}

	status = x1f4_type_e4fine
	    (lxtrap_data->link_o.fine, &type, linetext_over);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	status = x1f4_cast_node
	    (&lxtrap_data->link_v, &lxtrap_data->link_e, nodetype_data, type,
	     &linetext_over, output);
    } while (0);

    if (status) {
	if (leadnode_data) {
	    if (line_data->track) {
		if (linked) {
		    if (line_data->fold) {
			_libx1f4i0_lxtrap_free_args(line_data, NULL);
		    }

		    _x1f4_free_effect
			(NULL, linetext_data, NULL, &leadnode_data->link);
		}

		lxtrap_data->link_m.free
		    (lxtrap_data->link_m.data, line_data->track - 1);
	    }

	    lxtrap_data->link_m.free
		(lxtrap_data->link_m.data, leadnode_data);
	}
    }

    return status;
}
